/*
 * Copyright (c) 1994
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* $Id: route.c,v 1.3 1994/12/06 00:20:15 he Exp $ */

#include <stdio.h>
#include <sys/time.h>

#include <isode/pepsy/SNMP-types.h>
#include <isode/snmp/objects.h>
#include <isode/tailor.h>
#include <isode/dgram.h>
#include <isode/tsap.h>
#include <isode/internet.h>

#include "common.h"
#include "snmp_aux.h"
#include "utils.h"
#include "if.h"

extern int printf ();
extern void bcopy ();


static OID oid_ifDescr;
static OID oid_ifIndex;

static OID oid_ipRouteDest;
static OID oid_ipRouteIfIndex;
static OID oid_ipRouteNextHop;
static OID oid_ipRouteType;
static OID oid_ipRouteProto;
static OID oid_ipRouteMask;


struct var_info {
    caddr_t value;
    OS os;
    struct var_info *next;
};

struct if_info {
    caddr_t descr;
    int ix;
    struct if_info *next;
};

struct if_info *if_list = 0;


void
print_interface (fmt, ipRouteIfIndex)
    char *fmt;
    caddr_t ipRouteIfIndex;
{
    int ix;
    struct if_info *ip;
    char buf[16];

    ix = *((int*) ipRouteIfIndex);
    for (ip = if_list; ip; ip = ip -> next)
    {
	if (ip -> ix == ix) 
	{
	    print_if_descr (fmt, ip -> descr);
	    return;
	}
    }
    sprintf (buf, "if(%d)", ix);
    printf (fmt, buf);
}

void
process_if_descr (vbl)
    struct type_SNMP_VarBindList *vbl;
{
    struct type_SNMP_VarBindList *vp;
    struct type_SNMP_VarBind *v;
    caddr_t value;
    OI oi;
    OS os;
    caddr_t ifIndex, ifDescr;
    struct if_info *info;
    
    ifIndex = ifDescr = 0;

    for (vp = vbl; vp; vp = vp -> next) {
	v = vp -> VarBind;

	if ((oi = name2inst (v -> name)) == NULL ||
	    (os = oi -> oi_type -> ot_syntax) == NULL ||
	    (*os -> os_decode) (&value, v -> value) == NOTOK)
	{
	    vunknown (v -> value);
	}
	else
	{
	    if (oid_pfx_match (oid_ifIndex, v -> name))
		ifIndex = value;
	    if (oid_pfx_match (oid_ifDescr, v -> name))
		ifDescr = value;
	}
    }
    info = (struct if_info*) calloc(sizeof (struct if_info), 1);

    /* Layering violations, I know... */
    info -> ix = *((int*) ifIndex);
    info -> descr = ifDescr;
    info -> next = if_list;
    if_list = info;
}
	    

void
get_if_descs(ps, sd)
    PS ps;
    int sd;
{
    struct type_SNMP_Message *msg;

    msg = new_message (type_SNMP_PDUs_get__next__request, community);
    if (!msg)
	adios (NULLCP, "out of memory");
    
    oid_ifDescr = add_null_varbind (msg, "ifDescr");
    oid_ifIndex = add_null_varbind (msg, "ifIndex");

    getnext_process (msg, ps, sd, process_if_descr);
    return;
}


void
print_dest_and_mask (fmt, ipRouteDest, ipRouteMask)
    char *fmt;
    caddr_t ipRouteDest, ipRouteMask;
{
    struct sockaddr_in dest, mask;
    u_long m;
    int i;
    char mbuf[10], buf[28];
    
    bcopy ((char*) ipRouteDest, (char*) &dest,
	   sizeof (struct sockaddr_in));
    if (ipRouteMask) 
    {
	bcopy ((char*) ipRouteMask, (char*) &mask,
	       sizeof (struct sockaddr_in));

	for (m = mask.sin_addr.s_addr, i = 0; m & 0x80000000; i++, m <<= 1);
	if (m != 0) 
	    sprintf (mbuf, "0x%lx", mask.sin_addr.s_addr);
	else
	    sprintf (mbuf, "%d", i);
    }
    else
	sprintf (mbuf, "(none)");
	    
    sprintf(buf, "%s/%s", inet_ntoa(dest.sin_addr), mbuf);
    printf(fmt, buf);
}

void
print_next_hop (fmt, ipRouteNextHop)
    char *fmt;
    caddr_t ipRouteNextHop;
{
    char *cp = 0;
    struct hostent *hp;
    struct sockaddr_in gateway;
    static char line[MAXHOSTNAMELEN + 1];
    
    bcopy((char*)ipRouteNextHop, (char*) &gateway,
	  sizeof(struct sockaddr_in));
    
    if (!nflag) 
    {
	hp = gethostbyaddr((char *)&gateway.sin_addr, sizeof(struct in_addr),
			   AF_INET);
	if (hp) 
	    cp = hp -> h_name;
    }
    
    if (cp)
	strncpy(line, cp, sizeof(line) - 1);
    else 
	strcpy(line, inet_ntoa(gateway.sin_addr));
	
    printf(fmt, line);
}

void
print_flags(fmt, ipRouteType, ipRouteProto)
    char *fmt;
    caddr_t ipRouteType, ipRouteProto;
{
    int route_type, route_proto;
    static char buf[6];

    route_type = *((int*) ipRouteType);
    route_proto = *((int*) ipRouteProto);

    sprintf (buf, "U%s%s",	/* In use */
	    route_type == 4 ? "G" : "",	/* Indirect route */
	    route_proto == 4 ? "D" : ""); /* Route via ICMP */

    printf (fmt, buf);
}

static char *route_protos[] = {
    "Other", "Local/Static", "Via-SNMP", "Via-ICMP", "EGP", "GGP",
    "Hello", "RIP", "IS-IS", "ES-IS", "Cisco-IGRP", "BBN-SPF", "OSPF",
    "BGP"
};


void
print_route_proto(fmt, ipRouteProto)
    char *fmt;
    caddr_t ipRouteProto;
{
    
    enum_print (fmt, ipRouteProto, route_protos,
		sizeof route_protos / sizeof route_protos[0]);
}

void
process_rt_info (vbl)
    struct type_SNMP_VarBindList *vbl;
{
    struct type_SNMP_VarBindList *vp;
    struct type_SNMP_VarBind *v;
    struct var_info *vars = NULL;
    struct var_info *new_var, *vip;
    caddr_t value;
    OI oi;
    OS os;
    caddr_t ipRouteDest, ipRouteIfIndex, ipRouteNextHop;
    caddr_t ipRouteType, ipRouteProto, ipRouteMask;
    
    ipRouteDest = ipRouteIfIndex = ipRouteNextHop =
	ipRouteType = ipRouteProto = ipRouteMask = NULL;

    for (vp = vbl; vp; vp = vp -> next) {
	v = vp -> VarBind;

	if ((oi = name2inst (v -> name)) == NULL ||
	    (os = oi -> oi_type -> ot_syntax) == NULL ||
	    (*os -> os_decode) (&value, v -> value) == NOTOK)
	{
	    vunknown (v -> value);
	}
	else
	{
	    new_var = (struct var_info*) calloc(sizeof (struct var_info), 1);
	    new_var -> next = vars;
	    vars = new_var;

	    new_var -> value = value;
	    new_var -> os = os;

	    if (oid_pfx_match (oid_ipRouteDest, v -> name))
		ipRouteDest = value;
	    if (oid_pfx_match (oid_ipRouteIfIndex, v -> name))
		ipRouteIfIndex = value;
	    if (oid_pfx_match (oid_ipRouteNextHop, v -> name))
		ipRouteNextHop = value;
	    if (oid_pfx_match (oid_ipRouteType, v -> name))
		ipRouteType = value;
	    if (oid_pfx_match (oid_ipRouteProto, v -> name))
		ipRouteProto = value;
	    if (oid_pfx_match (oid_ipRouteMask, v -> name))
		ipRouteMask = value;
	}
    }	
    print_dest_and_mask ("%-21.21s ", ipRouteDest, ipRouteMask);
    print_next_hop ("%-22.22s ", ipRouteNextHop);
    print_flags("%-5.5s ", ipRouteType, ipRouteProto);
    print_interface("%-15s ", ipRouteIfIndex);
    print_route_proto("%s", ipRouteProto);

    printf("\n");

    for (vip = vars; vip; vip = vip -> next)
	(*vip -> os -> os_free) (vip -> value);
}

    
static char*
pfx_add (str)
    char *str;
{
    static char buf[100], tmbuf[100];
    int i;

    (void) strcpy (buf, str);
    for (i=0; i < prefix_len; i++) 
    {
	(void) sprintf (tmbuf, "%s.%d", buf, prefix[i]);
	(void) strcpy (buf, tmbuf);
    }
    return buf;
}

    
void
rheading ()
{

    printf("Routing tables\n");
    printf("%-21.21s %-22.22s %-5.5s %-15s %s\n",
	   "Destination (pfx/len)", "Gateway", "Flags",
	   "Interface", "Protocol");
}

void
routepr (ps, sd)
    PS ps;
    int sd;
{
    struct type_SNMP_Message *msg;

    msg = new_message (type_SNMP_PDUs_get__next__request, community);
    if (!msg)
	adios (NULLCP, "out of memory");
	
    get_if_descs(ps, sd);

    if (!hflag)
	rheading ();

    oid_ipRouteIfIndex = add_null_varbind (msg, pfx_add("ipRouteIfIndex"));
    oid_ipRouteNextHop = add_null_varbind (msg, pfx_add("ipRouteNextHop"));
    oid_ipRouteType    = add_null_varbind (msg, pfx_add("ipRouteType"));
    oid_ipRouteProto   = add_null_varbind (msg, pfx_add("ipRouteProto"));
    oid_ipRouteMask    = add_null_varbind (msg, pfx_add("ipRouteMask"));
    oid_ipRouteDest    = add_null_varbind (msg, pfx_add("ipRouteDest"));

    getnext_process (msg, ps, sd, process_rt_info);
    return;
}
