/*
 * Copyright (c) 1994
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* $Id: snmp_aux.c,v 1.3 1994/12/06 00:20:46 he Exp $ */

#include <stdio.h>

#include <isode/pepsy/SNMP-types.h>
#include <isode/snmp/objects.h>
#include <isode/tailor.h>
#include <isode/dgram.h>
#include <isode/tsap.h>
#include <isode/internet.h>

#include <sys/types.h>
#include <sys/time.h>

#include "common.h"
#include "snmp_aux.h"
#include "utils.h"

extern long random ();
extern int bcmp ();
extern void bzero ();
extern int gettimeofday ();
extern int select ();
extern int fprintf ();
extern int prnt_f ();
extern char* snmp_error ();


static OID do_add_null_varbind __P((struct type_SNMP_VarBind*, char*));
static OID do_add_varbind __P((struct type_SNMP_VarBind*, char*, char*));

static int watch = 0;


struct type_SNMP_Message*
new_message (msg_type, community)
    int msg_type;
    char *community;
{
    register struct type_SNMP_Message *msg;
    register struct type_SNMP_PDUs *pdu;
    register struct type_SNMP_PDU *parm;

    if ((msg = (struct type_SNMP_Message *) calloc (1, sizeof *msg)) == NULL)
	adios (NULLCP, "out of memory");
    
    msg -> version = int_SNMP_version_version__1;
    
    if ((msg -> community = str2qb (community, strlen (community), 1)) == NULL)
	adios (NULLCP, "out of memory");

    if ((pdu = (struct type_SNMP_PDUs *) calloc (1, sizeof *pdu)) == NULL)
	adios (NULLCP, "out of memory");

    msg -> data = pdu;
    pdu -> offset = msg_type;
    if ((parm = (struct type_SNMP_PDU *) calloc (1, sizeof *parm)) == NULL)
	adios (NULLCP, "out of memory");
    pdu -> un.get__request = parm;

    parm -> request__id = ((int) random ()) & 0x7fffffff;

    return msg;
}


OID
add_null_varbind (msg, varname)
    struct type_SNMP_Message *msg;
    char *varname;
{
    register struct type_SNMP_VarBindList **vp;
    register struct type_SNMP_VarBindList *bind;
    register struct type_SNMP_VarBind *v;
    OID oid;

    vp = &msg -> data -> un.get__request -> variable__bindings;
    
    if ((bind = (struct type_SNMP_VarBindList *) calloc (1, sizeof *bind))
	== NULL)
	adios (NULLCP, "out of memory");
    bind -> next = *vp;
    *vp = bind;

    if ((v = (struct type_SNMP_VarBind *) calloc (1, sizeof *v)) == NULL)
	adios (NULLCP, "out of memory");
    bind -> VarBind = v;
    
    if ((oid = do_add_null_varbind (v, varname)) == NULLOID)
    {
	free_SNMP_Message (msg);
	return NULL;
    }

    return oid;
}
	

OID
add_varbind (msg, varname, value)
    struct type_SNMP_Message *msg;
    char *varname;
    char *value;
{    
    register struct type_SNMP_VarBindList **vp;
    register struct type_SNMP_VarBindList *bind;
    register struct type_SNMP_VarBind *v;
    OID oid;

    vp = &msg -> data -> un.get__request -> variable__bindings;
    
    if ((bind = (struct type_SNMP_VarBindList *) calloc (1, sizeof *bind))
	== NULL)
	adios (NULLCP, "out of memory");
    bind -> next = *vp;
    *vp = bind;

    if ((v = (struct type_SNMP_VarBind *) calloc (1, sizeof *v)) == NULL)
	adios (NULLCP, "out of memory");
    bind -> VarBind = v;
    
    if ((oid = do_add_varbind (v, varname, value)) == NULLOID) 
    {
	free_SNMP_Message (msg);
	return NULL;
    }
    
    return oid;
}


static OID
do_add_null_varbind (v, varname)
    struct type_SNMP_VarBind *v;
    char *varname;
{

    return do_add_varbind (v, varname, NULLCP);
}    


static OID
do_add_varbind (v, varname, value_str)
    struct type_SNMP_VarBind *v;
    char *varname;
    char *value_str;
{
    int result;
    OS os;
    OI oi;
    OT ot;
    OID oid;
    caddr_t value;

    if ((oi = text2inst (varname)) == NULL) 
    {
	if ((oid = text2oid (varname)) == NULL) 
	{
	    advise (NULLCP, "unknown variable \"%s\"", varname);
	    return NULLOID;
	}
	ot = NULLOT;
    }
    else 
    {
	ot = oi -> oi_type;
	oid = NULLOID;
    }

    if ((v -> name = oid_cpy (oi ? oi -> oi_name : oid)) == NULLOID)
	adios (NULLCP, "out of memory");

    if (value_str == NULL) 
    {
	if ((v -> value = pe_alloc (PE_CLASS_UNIV, PE_FORM_PRIM,
				    PE_PRIM_NULL))
	    == NULLPE)
	    adios (NULLCP, "out of memory");
    }
    else 
    {
	if ((os = ot -> ot_syntax) == NULL) {
	    advise (NULLCP, "no syntax defined for object \"%s\"", varname);
	    return NULLOID;
	}
	
	if ((*os -> os_parse) (&value, value_str) == NOTOK) {
	    advise (NULLCP, "invalid value for variable \"%s\": \"%s\"",
		    varname, value_str);
	    return NULLOID;
	}
	result = (*os -> os_encode) (value, &v -> value);
	(*os -> os_free) (value);

	if (result == NOTOK) {
	    advise (NULLCP, "encoding error for variable \"%s\"", varname);
	    return NULLOID;
	}	
    }

    return oid_cpy (oi ? oi -> oi_name : oid);
}	

int
oid_pfx_match (pfx, oid)
    OID pfx, oid;
{
    int result;
    
    if (!oid || !pfx)
	return 0;

    result = bcmp ((char*) pfx -> oid_elements,
		   (char*) oid -> oid_elements,
		   pfx -> oid_nelem * sizeof pfx -> oid_elements[0]);
    return !result;
}

int
wait_for_action(sd)
    int sd;
{
    fd_set  rfds;
    int nfds, n;
#define MAXTIME (30 * 1000L)	/* 30 seconds, maximum timeout */
#define MINTIME 1		/* 1 millisecond, minimum timeout */
    static u_long timeout = 2 * 1000L;
    static u_long lasttime;
    static u_long timenow;
    static long measure, avgrtt, devrtt;
    struct timeval tvs;
    
    FD_ZERO (&rfds);
    nfds = sd + 1;
    FD_SET (sd, &rfds);
    
    (void) gettimeofday (&tvs, (struct timezone *) 0);
    timenow = tvs.tv_sec * 1000L + tvs.tv_usec / 1000L;
    lasttime = timenow;

    tvs.tv_sec = timeout / 1000L, tvs.tv_usec = (timeout % 1000) * 1000L;

    if ((n = select (nfds, &rfds, NULLFD, NULLFD, &tvs)) < 0) 
    {
	advise ("failed", "select");
	return -1;
    }
    
    (void) gettimeofday (&tvs, (struct timezone *) 0);
    timenow = tvs.tv_sec * 1000L + tvs.tv_usec / 1000L;

    if (n == 0)			/* Timeout? */
    {
	avgrtt <<= 1;
	timeout = ((avgrtt >> 2) + devrtt) /* >> 1 */;

	if (timeout > MAXTIME)
	    timeout = MAXTIME;
	if (debug)
	    fprintf (stderr, "timeout, adjusted timeout to %g seconds\n",
		     timeout / 1000.0);
	return -1;
    }
    else
    {
	measure = timenow - lasttime;
	if (debug)
	    fprintf (stderr, "time now %lu, xmit-time %lu, RTT %lu\n",
		     timenow, lasttime, measure);
	
	measure -= (avgrtt >> 3);
	avgrtt += measure;
	if (measure < 0)
	    measure = -measure;
	measure -= (devrtt >> 2);
	devrtt += measure;
	timeout = ((avgrtt >> 2) + devrtt) /* >> 1 */;

	if (timeout > MAXTIME)
	    timeout = MAXTIME;
	else
	    if (timeout < MINTIME)
		timeout = MINTIME;

	if (debug)
	    fprintf (stderr, "adjusted timeout to %g seconds(0)\n",
		     timeout / 1000.0);
    }
    return 0;
}    


#define MAX_RETRY 10
    
int
getnext_process (msg, ps, sd, process_vars)
    struct type_SNMP_Message *msg;
    PS ps;
    int sd;
    void (*process_vars)();
{
    int request_id;
    OID first_oid;
    PE pe;
    struct type_SNMP_VarBindList *vp;
    register struct type_SNMP_PDU *pdu;
    struct type_SNMP_VarBind *v;
    int retry = 0;
    
    request_id = msg -> data -> un.get__response -> request__id;

    if ((first_oid = oid_cpy (msg -> data -> un.get__next__request ->
			      variable__bindings -> VarBind -> name))
	== NULLOID) 
	adios (NULLCP, "out of memory");

 again:
    pe = NULLPE;

    if (encode_SNMP_Message (&pe, 1, 0, NULLCP, msg) != NOTOK) {
	if (watch)
	    (void) print_SNMP_Message (pe, 1, NULLIP, NULLVP, NULLCP);

	if (pe2ps (ps, pe) == NOTOK) {
	    advise (NULLCP, "pe2ps: %s", ps_error (ps -> ps_errno));
	    goto out;
	}
    }
    else
	advise (NULLCP, "encode_SNMP_Message: %s", PY_pepy);
    
 try_again:
    if (wait_for_action(sd) != 0)
	if (++retry < MAX_RETRY)
	    goto again;
	else
	    adios (NULLCP, "SNMP timeout, max retries");

    if (pe)
	pe_free (pe);
    pe = NULLPE;

    free_SNMP_Message (msg);
    msg = NULL;

    if ((pe = ps2pe (ps)) == NULLPE) 
    {
	advise (NULLCP, "ps2pe: %s", ps_error (ps -> ps_errno));
	goto out;
    }

    if (decode_SNMP_Message (pe, 1, NULLIP, NULLVP, &msg) == NOTOK) {
	advise (NULLCP, "decode_SNMP_Message: %s", PY_pepy);
	goto out;
    }

    if (watch)
	(void) print_SNMP_Message (pe, 1, NULLIP, NULLVP, NULLCP);
    
    if (msg -> data -> offset != type_SNMP_PDUs_get__response) {
	advise (NULLCP, "unexpected message type %d",
		msg -> data -> offset);
	goto out;
    }

    if ((pdu = msg -> data -> un.get__response) -> request__id
	    != request_id) {
	if (debug)
	    advise (NULLCP, "request-id mismatch (got %d, wanted %d)",
		    pdu -> request__id, request_id);
	retry = 0;
	goto try_again;
    }
    
    if (pdu -> error__status != int_SNMP_error__status_noError) {
	if (pdu -> error__status != int_SNMP_error__status_noSuchName)
	    advise (NULLCP, "%s at position %d\n",
		     snmp_error (pdu -> error__status), pdu -> error__index);
	goto out;
    }

    v = pdu -> variable__bindings -> VarBind;
    if (! oid_pfx_match (first_oid, v -> name))
	goto out;
    
    (*process_vars) (pdu -> variable__bindings);

    if (pe)
	pe_free (pe);
    msg -> data -> offset = type_SNMP_PDUs_get__next__request;
    request_id = ++pdu -> request__id;

    for (vp = pdu -> variable__bindings; vp; vp = vp -> next) {
	v = vp -> VarBind;

	if ((v -> value = pe_alloc (PE_CLASS_UNIV, PE_FORM_PRIM,
				    PE_PRIM_NULL))
	    == NULLPE)
	    adios (NULLCP, "out of memory");
    }
    
    goto again;
    
 out:
    if (first_oid)
	oid_free (first_oid);
    if (pe)
	pe_free (pe);
    if (msg)
	free_SNMP_Message (msg);

    return OK;
}

int
get_process (msg, ps, sd, process_vars)
    struct type_SNMP_Message *msg;
    PS ps;
    int sd;
    void (*process_vars)();
{
    int request_id;
    OID first_oid;
    PE pe;
    register struct type_SNMP_PDU *pdu;
    struct type_SNMP_VarBind *v;
    int retry = 0;
    
    request_id = msg -> data -> un.get__response -> request__id;

    if ((first_oid = oid_cpy (msg -> data -> un.get__next__request ->
			      variable__bindings -> VarBind -> name))
	== NULLOID) 
	adios (NULLCP, "out of memory");

  again:
    pe = NULLPE;

    if (encode_SNMP_Message (&pe, 1, 0, NULLCP, msg) != NOTOK) {
	if (watch)
	    (void) print_SNMP_Message (pe, 1, NULLIP, NULLVP, NULLCP);

	if (pe2ps (ps, pe) == NOTOK) {
	    advise (NULLCP, "pe2ps: %s", ps_error (ps -> ps_errno));
	    goto out;
	}
    }
    else
	advise (NULLCP, "encode_SNMP_Message: %s", PY_pepy);
    
 try_again:
    if (wait_for_action(sd) != 0)
	if (++retry < MAX_RETRY)
	    goto again;
	else
	    adios (NULLCP, "SNMP timeout, max retries");

    if (pe)
	pe_free (pe);
    pe = NULLPE;

    free_SNMP_Message (msg);
    msg = NULL;

    if ((pe = ps2pe (ps)) == NULLPE) 
    {
	advise (NULLCP, "ps2pe: %s", ps_error (ps -> ps_errno));
	goto out;
    }

    if (decode_SNMP_Message (pe, 1, NULLIP, NULLVP, &msg) == NOTOK) {
	advise (NULLCP, "decode_SNMP_Message: %s", PY_pepy);
	goto out;
    }

    if (watch)
	(void) print_SNMP_Message (pe, 1, NULLIP, NULLVP, NULLCP);
    
    if (msg -> data -> offset != type_SNMP_PDUs_get__response) {
	advise (NULLCP, "unexpected message type %d",
		msg -> data -> offset);
	goto out;
    }

    if ((pdu = msg -> data -> un.get__response) -> request__id
	!= request_id) {
	if (debug)
	    advise (NULLCP, "request-id mismatch (got %d, wanted %d)",
		    pdu -> request__id, request_id);
	retry = 0;
	goto try_again;
    }
    
    if (pdu -> error__status != int_SNMP_error__status_noError) {
	if (pdu -> error__status != int_SNMP_error__status_noSuchName)
	    advise (NULLCP, "%s at position %d\n",
		     snmp_error (pdu -> error__status), pdu -> error__index);
	goto out;
    }

    v = pdu -> variable__bindings -> VarBind;
    if (! oid_pfx_match (first_oid, v -> name))
	goto out;
    
    process_vars (pdu -> variable__bindings);

    if (pe)
	pe_free (pe);
    
 out:
    if (first_oid)
	oid_free (first_oid);
    if (pe)
	pe_free (pe);
    if (msg)
	free_SNMP_Message (msg);

    return OK;
}

static OID save_oid;
static int save_value;

void
process_int_var (vbl)
    struct type_SNMP_VarBindList *vbl;
{
    struct type_SNMP_VarBindList *vp;
    struct type_SNMP_VarBind *v;
    caddr_t value = NULL;
    OI oi;
    OS os;

    for (vp = vbl; vp; vp = vp -> next) {
	v = vp -> VarBind;

	if ((oi = name2inst (v -> name)) == NULL ||
	    (os = oi -> oi_type -> ot_syntax) == NULL ||
	    (*os -> os_decode) (&value, v -> value) == NOTOK)
	{
	    vunknown (v -> value);
	}
	else
	{
	    if (!oid_pfx_match (save_oid, v -> name))
	    {
		advise(NULLCP, "unknown variable in response %s",
		       oid2ode (v -> name));
		return;
	    }
	    save_value = *((int*) value);
	    (*os -> os_free) (value);
	}
    }
}


int
get_int_var (ps, sd, varname)
    PS ps;
    int sd;
    char *varname;
{
    struct type_SNMP_Message *msg;
    
    msg = new_message (type_SNMP_PDUs_get__request, community);
    save_oid = add_null_varbind (msg, varname);
    get_process (msg, ps, sd, process_int_var);
    free (save_oid);

    return save_value;
}

