/*
 * Copyright (c) 1994
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* $Id: snmpnetstat.c,v 1.3 1994/12/06 00:21:24 he Exp $ */

/*
 * Reimplementation of cmu-snmp/apps/snmpnetstat + my extensions
 * using the ISODE libraries
 */

#include <stdio.h>

#include <isode/pepsy/SNMP-types.h>
#include <isode/snmp/objects.h>
#include <isode/tailor.h>
#include <isode/dgram.h>
#include <isode/tsap.h>
#include <isode/internet.h>

#include <sys/types.h>
#include <sys/time.h>
#include <ctype.h>

#include "common.h"
#include "transport.h"
#include "route.h"
#include "if.h"
#include "utils.h"

extern int loadobjects ();
extern int srandom ();
extern time_t time ();
extern int atoi ();

extern int ps_len_strategy;

static int parse_prefix __P((char *));

char	*community = "public";
char	*myname = "snmpnetstat";

int	iflag;
int	rflag;
int	mflag;
int	Mflag;
int	nflag;
int	hflag;
int	Bflag;
int	Cflag;

int	prefix[4];
int	prefix_len = 0;

int	debug = 0;

static char	*agent;
static PS	ps;
static int	sd;
static int	interval;
static char *usage = "usage: %s agent community -[irBCmMnh] prefix/interval";
static char *defs = NULLCP;


static void
initialize()
{

    ps_len_strategy = PS_LEN_LONG;

    if (loadobjects (defs) == NOTOK)
	adios (NULLCP, "loadobjects: %s", PY_pepy);

    (void) srandom ((int) time ((long *) 0));
}
     

static void
arg_parse (argc, vec)
    int argc;
    char **vec;
{
    char *ap;
    
    if ((myname = rindex (*vec, '/')) != NULL)
	myname++;
    if (myname == NULL || *myname == NULL)
	myname = *vec;
    
    vec++;
    argc--;
    
    if (argc--)
	agent = *vec++;
    if (argc--)
	community = *vec++;
    else 
	adios (NULLCP, usage, myname);

    for (; argc > 0 && **vec == '-'; vec++, argc--) 
    {
	for (ap = *vec; *++ap; )
	    switch (*ap) 
	    {
	    case 'i':
		iflag++;
		break;

	    case 'r':
		rflag++;
		break;

	    case 'm':
		mflag++;
		break;

	    case 'M':
		Mflag++;
		break;

	    case 'n':
		nflag++;
		break;

	    case 'h':
		hflag++;
		break;

	    case 'B':
		Bflag++;
		break;

	    case 'C':
		Cflag++;
		break;

	    default:
		adios (NULLCP, usage, myname);
	    }
    }
    
    
    if (rflag && argc > 0 && isdigit (**vec)) 
	if (parse_prefix (*vec) != 0)
	    adios (NULLCP, usage, myname);

    if (!rflag && argc > 0 && isdigit (**vec)) 
    {
	interval = atoi(*vec);
	if (interval <= 0)
	    adios (NULLCP, usage, myname);
	vec++;
	argc--;
	iflag++;
    }
}


static int
parse_prefix(pp)
    char *pp;
{
    char *dot;
    int i;
    
    dot = pp;
    do
    {
	if ((dot = index (pp, '.')) != NULL )
	    *dot = '\0';
	i = atoi (pp);
	if (i >= 0 && i <= 255 && prefix_len <= 4)
	    prefix[prefix_len++] = i;
	else
	    return -1;
	pp = dot + 1;
    }while (dot != NULL);
    
    for(i = prefix_len - 1; i >= 0 && prefix[i] == 0; i--, prefix_len--);

    if (prefix_len == 0) {
	prefix[0] = 0;
	prefix_len = 1;
    }
#ifdef notdef
    printf("prefix is: ");
    for(i=0; i<prefix_len; i++)
    {
	printf("%d",prefix[i]);
	if ( i != prefix_len-1 )
	    putchar('.');
    }
    putchar('\n');
#endif
    return 0;
}

int
main (argc, argv)
    int argc;
    char **argv;
{

    initialize ();

    arg_parse (argc, argv);
    
    transport_init(agent, &ps, &sd);
    
    if (iflag || Cflag || Bflag)
    {
	intpr (interval, ps, sd);
	exit (0);
    }
    if (rflag)
	routepr(ps, sd);
    
    exit(0);
}
