/*
 * Copyright (c) 1994
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* $Id: transport.c,v 1.3 1994/12/06 00:21:56 he Exp $ */

#include <stdio.h>

#include <isode/pepsy/SNMP-types.h>
#include <isode/snmp/objects.h>
#include <isode/tailor.h>
#include <isode/dgram.h>
#include <isode/tsap.h>
#include <isode/internet.h>

#include "common.h"
#include "transport.h"
#include "utils.h"

extern void bzero ();
extern void bcopy ();


void
transport_init (agent, psp, sdp)
    char *agent;
    PS *psp;
    int *sdp;
{
    struct sockaddr_in in_socket;
    register struct sockaddr_in *isock = &in_socket;
    register struct hostent *hp;
    register struct servent *sp;
    static struct TSAPaddr snmp_ta;

    register struct TSAPaddr *ta = &snmp_ta, *tz;
    register struct NSAPaddr *na = ta -> ta_addrs;
    struct sockaddr_in lo_socket;
    register struct sockaddr_in *lsock = &lo_socket;

    if ((sp = getservbyname ("snmp", "udp")) == NULL)
	advise (NULLCP, "udp/snmp: unknown service");
    
    bzero ((char *)ta, sizeof *ta);
    
    na -> na_stack = NA_TCP;
    na -> na_community = ts_comm_tcp_default;
    (void) strncpy (na -> na_domain, getlocalhost (),
		    sizeof na -> na_domain - 1);
    na -> na_port = sp ? sp -> s_port : htons ((u_short) 161);
    na -> na_tset = NA_TSET_UDP;
    ta -> ta_naddr = 1;
    
    if ((hp = gethostbystring (agent)) != NULL) {
	inaddr_copy (hp, isock);
	(void) strncpy (na -> na_domain,
			inet_ntoa(isock -> sin_addr),
			sizeof na -> na_domain - 1);
    }
    else
	if ((tz = str2taddr (agent)) && tz -> ta_naddr > 0) {
	    *ta = *tz;	/* struct copy */
	    if (na -> na_stack == NA_TCP) {
		if (na -> na_port == 0)
		    na -> na_port = sp ? sp -> s_port : htons ((u_short) 161);
		na -> na_tset = NA_TSET_UDP;
	    }
	}
	else
	    adios (NULLCP, "%s: unknown host", agent);
    
    if (na -> na_stack != NA_TCP) 
	adios (NULLCP, "only supports UDP/IP transport");
    

    bzero ((char *) lsock, sizeof *lsock);

    if ((hp = gethostbystring (agent = getlocalhost ())) == NULL)
	adios (NULLCP, "%s: unknown host", agent);
    lsock -> sin_family = hp -> h_addrtype;
    inaddr_copy (hp, lsock);
    
    if ((*sdp = start_udp_client (lsock, 0, 0, 0)) == NOTOK)
	adios ("failed", "start_udp_client");
    
    bzero ((char *) isock, sizeof *isock);
    if ((hp = gethostbystring (na -> na_domain)) == NULL)
      adios (NULLCP, "%s: unknown host", na -> na_domain);
    isock -> sin_family = hp -> h_addrtype;
    isock -> sin_port = na -> na_port;
    inaddr_copy (hp, isock);
    
    if (join_udp_server (*sdp, isock) == NOTOK)
	adios ("failed", "join_udp_server");

    if ((*psp = ps_alloc (dg_open)) == NULLPS
	|| dg_setup (*psp, *sdp, MAXDGRAM, read_udp_socket,
		     write_udp_socket, check_udp_socket) == NOTOK) {
	if (*psp == NULLPS)
	    adios (NULLCP, "ps_alloc: out of memory");
	else
	    adios (NULLCP, "dg_setup: %s", ps_error ((*psp) -> ps_errno));
    }
}
