/*
 * Copyright (c) 1994
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* $Id: utils.c,v 1.3 1994/12/06 00:22:27 he Exp $ */

#include <stdio.h>
#include <varargs.h>
#include <sys/types.h>
#include <sys/cdefs.h>

#include "common.h"
#include "utils.h"

extern int fflush ();
extern int fprintf ();
extern int fputs ();
extern int fputc ();
extern int printf ();

extern void asprintf ();

static void _advise();


void
adios(va_alist)
    va_dcl
{
    va_list ap;
    
    va_start (ap);
    
    _advise (ap);
    
    va_end (ap);
    
    _exit (1);
}


void
advise(va_alist)
    va_dcl
{
    va_list ap;
    
    va_start (ap);
    
    _advise (ap);
    
    va_end (ap);
}


static void
_advise(ap)
    va_list ap;
{
    char buffer[BUFSIZ];

    asprintf (buffer, ap);

    (void) fflush (stdout);
    
    fprintf (stderr, "%s: ", myname);
    (void) fputs (buffer, stderr);
    (void) fputc ('\n', stderr);
    
    (void) fflush (stderr);
}

static char *errors[] = {
    "noError", "tooBig", "noSuchName", "badValue", "readOnly", "genErr"
};

char *
snmp_error (i)
     int i;
{
    static char buffer[BUFSIZ];
    
    if (0 < i && i < sizeof errors / sizeof errors[0])
	return errors[i];
    (void) sprintf (buffer, "error %d", i);
    
    return buffer;
}


void
enum_print (fmt, value, strings, nstrings)
    char *fmt;
    caddr_t value;
    char **strings;
    int nstrings;
{
    int i = *((int*) value);
    static char buf[40];

    if (i <= 0 || i > nstrings)
    {
	sprintf (buf, "unknown(%d)", i);
	printf (fmt, buf);
    }
    else
	printf (fmt, strings[i-1]);
}
